/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mediator.algebra;

import java.util.ArrayList;

import org.xquark.mediator.decomposer.GetVariablePathsVisitor;
import org.xquark.mediator.decomposer.IsolateWhereVisitor;
import org.xquark.mediator.plan.ExecutionPlan;
import org.xquark.mediator.plan.OpSourceEval;
import org.xquark.mediator.plan.Operator;
import org.xquark.mediator.runtime.MediatorException;
import org.xquark.xquery.parser.FLWRExpression;
import org.xquark.xquery.parser.ParenthesizedExpression;
import org.xquark.xquery.parser.Variable;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.parser.XQueryExpression;
import org.xquark.xquery.typing.QTypeAtom;
import org.xquark.xquery.typing.TypeException;

/**
 * This class represent the terminal leaf of the dependancy tree. It is
 * generated after atomization (and is resulted from the transformation of
 * DepSource that disappear).
 *  
 */
public class AlgSourceEval extends Algebra {
    // **********************************************************************
    // * VERSIONING
    // **********************************************************************
    private static final String RCSRevision = "$Revision: 1.15 $";

    private static final String RCSName = "$Name: XQ1_1_1 $";

    // **********************************************************************
    // * CLASS VARIABLES
    // **********************************************************************
    //private boolean hasIdentifier = false;
    // ************************************************************************
    // * INITIALIZATION
    // ************************************************************************
    /**
     * Create a node from a variable and an expression.
     * 
     * To create the node, we take the original expression, and we extract
     * everything relative to variable given by the leaf. from that, we create a
     * new flower expression that will be associated to this DepNodeSource, and
     * the old XQueryExpression that is referenced by it's DepNode is
     * simplified. (bon, en gros, a sert a faire une optimisation en faisant
     * redescendre les restrictions (et plus tard les projection) vers les
     * sources.
     * 
     * @param typevisitor
     * @param origexp
     * @param varleaf
     * @param depmanager
     */
    public AlgSourceEval(XQueryExpression origexp, ArrayList variables, AlgebraManager depmanager, boolean islet) throws XQueryException {
        // to satisfy the super constructor, but expression and variable will be
        // changed in following step.
        super(origexp, variables, depmanager, islet);
    }

    public AlgSourceEval(XQueryExpression origexp, Variable varleaf, AlgebraManager depmanager, boolean islet) throws XQueryException {
        // to satisfy the super constructor, but expression and variable will be
        // changed in following step.
        super(origexp, varleaf, depmanager, islet);
    }

    // ******************************************************************
    // * METHODS
    // ******************************************************************

    public void ParentHasIdentifier() {
        hasIdentifier = true;
    }

    /**
     * Create the FLWRExpression like this : - F : is the variable varleaf that
     * we consider with no changes (even the expression associated to it is kept ; -
     * L : there is no Let at this point ; - W : is the major part, we must scan
     * the origexp to see what to keep - R : for now, we return just the
     * variable, but later, when we will care about prjection, we will make good
     * LocatedExpression with it.
     * 
     * @param origexp
     * @param varleaf
     */

    private void createFLWRExpression() throws XQueryException {
        if (!(expression instanceof FLWRExpression))
            throw new MediatorException("AlgSourceEval : " + expression + " is not a FLWR.");

        try {
            FLWRExpression origflwr = (FLWRExpression) expression;

            // Variables
            ArrayList newvariables = (ArrayList) variables.clone();
            newvariables.addAll(this.parent.getVarsDependingOn());
            newvariables.addAll(this.getVarsDependingOn());
            newvariables.addAll(this.getVarsWhereOn());

            // Where
            XQueryExpression newWhereClause = null;
            if (origflwr.getWhereClause() != null) {
                newWhereClause = (XQueryExpression) origflwr.getWhereClause().clone(null);
                IsolateWhereVisitor visitor = new IsolateWhereVisitor(newvariables);
                newWhereClause.accept(visitor);
                newWhereClause = visitor.getIsolatedExpression();
                // erase isolated where clause from original where clause
                ((FLWRExpression) expression).setWhereClause(visitor.getRemainingExpression());
            }

            // update paths because where clause is not calculate in flwr
            if (newWhereClause != null) {
                GetVariablePathsVisitor gvpv = new GetVariablePathsVisitor(true);
                newWhereClause.accept(gvpv);
                ArrayList tmplist = gvpv.getPaths();
                if (tmplist != null)
                    for (int i = 0; i < tmplist.size(); i++)
                        if (!this.algManager.getPathsList().contains(tmplist.get(i).toString()))
                            paths.add(tmplist.get(i));
            }
            // make return expression
            FLWRExpression newFlwr = null;
            if (newWhereClause != null && newWhereClause.getQType() == null)
                newWhereClause.setQType(new QTypeAtom(this.algManager.getPlan().getSchemaManager().getBooleanType()));
            newFlwr = new FLWRExpression(variables, null, newWhereClause, new ParenthesizedExpression(null, expression.getParentModule()), expression.getParentModule());
            newFlwr.setParentModule(origflwr.getParentModule());
            expression = newFlwr;
        } catch (TypeException te) {
            throw new MediatorException("AlgSourceEval.createFLWRExpression : " + te.getMessage(), te);
        } catch (CloneNotSupportedException e) {
            throw new MediatorException("AlgSourceEval.createFLWRExpression : " + e.getMessage(), e);
        } catch (XQueryException e) {
            throw new MediatorException("AlgSourceEval.createFLWRExpression : " + e.getMessage(), e);
        }
    }

    // ******************************************************************
    // * INFORMATION
    // ******************************************************************
    /**
     * See what sources are associated to this DepNodeSources
     */

    // ******************************************************************
    // * CREATE EXECUTION PLAN
    // ******************************************************************
    /**
     * 
     * @param plan
     * @return
     */
    public Operator createOperator(ExecutionPlan plan) throws XQueryException {
        createFLWRExpression();
        Operator algebra = new OpSourceEval(plan, expression, paths, hasIdentifier);
        algebra.isLet(islet);
        return algebra;
    }

    public Object clone() throws CloneNotSupportedException {
        AlgSourceEval newobj = (AlgSourceEval) super.clone();
        newobj.hasIdentifier = this.hasIdentifier;
        return newobj;
    }

    // ******************************************************************
    // * GET/SET METHODS
    // ******************************************************************

}